@@ -8,19 +8,20 @@ Synopsis:
     for d in vnlog.vnlog(f):
         print(d['time'],d['height'])
 
-Vnlog is simple, and you don't NEED a parser to read it, but this makes it a bit
-nicer.
+Vnlog is simple, and you don't NEED a parser to read it, but this library makes
+it a bit nicer.
 
 This module provides three different ways to parse vnlog
 
 1. slurp the whole thing into a numpy array: the slurp() function. Basic usage:
 
    import vnlog
-   log_numpy_array,list_keys,dict_key_index = \
+   arr,list_keys,dict_key_index = \
         vnlog.slurp(filename_or_fileobject)
 
    This parses out the legend, and then calls numpy.loadtxt(). Null data values
-   ('-') are not supported
+   ('-') are not supported at this time. A structured dtype can be passed-in to
+   read non-numerical data. See the docstring for vnlog.slurp() for details
 
 2. Iterate through the records: vnlog class, used as an iterator. Basic usage:
 
@@ -30,7 +31,6 @@ This module provides three different ways to parse vnlog
 
    Null data values are represented as None
 
-
 3. Parse incoming lines individually: vnlog class, using the parse() method.
    Basic usage:
 
@@ -124,8 +124,8 @@ class vnlog:
         self._values_dict = None
 
         if not hasattr(self, 're_hard_comment'):
-            self.re_hard_comment = re.compile('^\s*(?:#[#!]|#\s*$|$)')
-            self.re_soft_comment = re.compile('^\s*#\s*(.*?)\s*$')
+            self.re_hard_comment = re.compile(r'^\s*(?:#[#!]|#\s*$|$)')
+            self.re_soft_comment = re.compile(r'^\s*#\s*(.*?)\s*$')
 
         if self.re_hard_comment.match(l):
             # empty line or hard comment.
@@ -223,26 +223,57 @@ class vnlog:
     next = __next__
 
 
-def _slurp(f):
+def _slurp(f,
+           *,
+           dtype = None):
     r'''Reads a whole vnlog into memory
 
     This is an internal function. The argument is a file object, not a filename.
 
-    Returns a tuple (log_numpy_array, list_keys, dict_key_index)
+    See the docs for slurp() for details
 
     '''
     import numpy as np
 
 
-    # Stripped down numpysane.atleast_dims(). I don't want to introduce that
-    # dependency
-    def atleast_dims(x, d):
-        if d >= 0: raise Exception("Requires d<0")
-        need_ndim = -d
-        if x.ndim >= need_ndim:
-            return x
-        num_new_axes = need_ndim-x.ndim
-        return x[ (np.newaxis,)*(num_new_axes) ]
+    # Expands the fields in a dtype into a flat list of names. For vnlog
+    # purposes this doesn't support multiple levels of fields and it doesn't
+    # support unnamed fields. It DOES support (require!) compound elements with
+    # whitespace-separated field names, such as 'x y z' for a shape-(3,) field.
+    #
+    # This function is an analogue of field_type_grow_recursive() in
+    # https://github.com/numpy/numpy/blob/9815c16f449e12915ef35a8255329ba26dacd5c0/numpy/core/src/multiarray/textreading/field_types.c#L95
+    def field_names_in_dtype(dtype,
+                             split_name = None,
+                             name       = None):
+
+        if dtype.subdtype is not None:
+            if split_name is None:
+                raise Exception("only structured dtypes with named fields are supported")
+            size = np.prod(dtype.shape)
+            if size != len(split_name):
+                raise Exception(f'Field "{name}" has {len(split_name)} elements, but the dtype has it associated with a field of shape {dtype.shape} with {size} elements. The sizes MUST match')
+            yield from split_name
+            return
+
+        if dtype.fields is not None:
+            if split_name is not None:
+                raise("structured dtype with nested fields unsupported")
+            for name1 in dtype.names:
+                tup = dtype.fields[name1]
+                field_descr = tup[0]
+
+                yield from field_names_in_dtype(field_descr,
+                                                name       = name1,
+                                                split_name = name1.split(),)
+            return
+
+        if split_name is None:
+            raise Exception("structured dtype with unnamed fields unsupported")
+        if len(split_name) != 1:
+            raise Exception(f"Field '{name}' is a scalar so it may not contain whitespace in its name")
+        yield split_name[0]
+
 
 
     parser = vnlog()
@@ -260,33 +291,148 @@ def _slurp(f):
     for i in range(len(keys)):
         dict_key_index[keys[i]] = i
 
-    return                               \
-        atleast_dims(np.loadtxt(f), -2), \
-        keys,                            \
-        dict_key_index
+
+    if dtype is None or \
+       not isinstance(dtype, np.dtype) or \
+       ( dtype.fields is None and \
+         dtype.subdtype is None ):
+        return \
+            ( np.loadtxt(f, ndmin=2, dtype=dtype),
+              keys,
+              dict_key_index )
 
 
-def slurp(f):
+    # We have a dtype. We parse out the field names from it, map those to
+    # columns in the input (from the vnl legend that we just parsed), and
+    # load everything with np.loadtxt()
+
+    names_dtype = list(field_names_in_dtype(dtype))
+
+    # We have input fields in the vnl represented in:
+    # - keys
+    # - dict_key_index
+    #
+    # We have output fields represented in:
+    # - names_dtype
+    #
+    # Each element of 'names_dtype' corresponds to each output field, in
+    # order. 'names_dtype' are names of these input fields, which must match
+    # the input names given in 'keys'.
+    Ncols_output = len(names_dtype)
+
+    usecols = [None] * Ncols_output
+
+    for i_out in range(Ncols_output):
+        name_dtype = names_dtype[i_out]
+        try:
+            i_in = dict_key_index[name_dtype]
+        except:
+            raise Exception(f"The given dtype contains field {name_dtype=} but this doesn't appear in the vnlog columns {keys=}")
+        usecols[i_out] = i_in
+
+    return \
+        np.loadtxt(f,
+                   ndmin   = 1,
+                   dtype   = dtype,
+                   usecols = usecols)
+
+
+
+def slurp(f,
+          *,
+          dtype = None):
     r'''Reads a whole vnlog into memory
 
-    Synopsis:
+SYNOPSIS
 
-        import vnlog
-        log_numpy_array,list_keys,dict_key_index = \
-             vnlog.slurp(filename_or_fileobject)
+    import vnlog
 
-        This parses out the legend, and then calls numpy.loadtxt(). Null data
-        values ('-') are not supported.
+    ### Read numerical data into arr
+    arr,list_keys,dict_key_index = \
+         vnlog.slurp(filename_or_fileobject)
 
-    Returns a tuple (log_numpy_array, list_keys, dict_key_index)
+    ### Read disparate, partly-numerical data using a structured dtype
+    # Let's say "data.vnl" contains:
+
+      #  image   x y z temperature
+      image1.png 1 2 5 34
+      image2.png 3 4 1 35
+
+    dtype = np.dtype([ ('image',       'U16'),
+                       ('x y z',       int, (3,)),
+                       ('temperature', float), ])
+
+    arr = vnlog.slurp("data.vnl", dtype=dtype)
+
+    print(arr['image'])
+    ---> array(['image1.png', 'image2.png'], dtype='<U16')
+
+    print(arr['x y z'])
+    ---> array([[1, 2, 5],
+                [3, 4, 1]])
+
+    print(arr['temperature'])
+    ---> array([34., 35.])
+
+This function is primarily a wrapper around numpy.loadtxt(), which does most of
+the work. Null data values ('-') are not supported at this time.
+
+A dtype can be given in a keyword argument. If this is a base type (something
+like 'float' or 'np.int8'), the returned array will be composed entirely of
+values of that type.
+
+If this is a structured dtype (like the one in the SYNOPSIS above), a structured
+array will be returned. Some notes about this behavior:
+
+- The given structured dtype defines both how to organize the data, and which
+  data to extract. So it can be used to read in only a subset of the available
+  columns. In the sample above I could have omitted the 'temperature' column,
+  for instance
+
+- Sub-arrays are allowed. In the example I could say either
+
+    dtype = np.dtype([ ('image',       'U16'),
+                       ('x y z',       int, (3,)),
+                       ('temperature', float), ])
+
+  or
+
+    dtype = np.dtype([ ('image',       'U16'),
+                       ('x',           int),
+                       ('y',           int),
+                       ('z',           int),
+                       ('temperature', float), ])
+
+  The latter would read x,y,z into separate, individual arrays. Sometime we want
+  this, sometimes not.
+
+- Nested structured dtypes are not allowed. Fields inside other fields are not
+  supported, since it's not clear how to map that to a flat vnlog legend
+
+- If a structured dtype is given we return the array only, since the field names
+  are already available in the dtype
+
+ARGUMENTS
+
+- f: a filename or a readable Python "file" object. We read this until the end
+
+- dtype: an optional dtype for the ouput array. May be a structured dtype
+
+RETURN VALUE
+
+- If no dtype is given or a simple dtype is given:
+  Returns a tuple (arr, list_keys, dict_key_index)
+
+- If a structured dtype is given:
+  Returns arr
 
     '''
 
     if type(f) is str:
         with open(f, 'r') as fh:
-            return _slurp(fh)
+            return _slurp(fh, dtype=dtype)
     else:
-        return _slurp(f)
+        return _slurp(f, dtype=dtype)
 
 
 
